import numpy as np
import transforms3d as t3d
from pyquaternion import Quaternion


class Transform:
    def __init__(self, value=None):
        self.matrix = np.eye(4)

        if isinstance(value, Transform):
            self.matrix = np.array(value)

        elif isinstance(value, Quaternion):
            self.rotation = value.rotation_matrix

        elif value is not None:
            value = np.array(value)
            if value.shape == (4, 4):
                self.matrix = value

            if value.shape == (16,):
                self.matrix = np.array(value).reshape((4, 4))

            elif value.shape == (3, 4):
                self.matrix[:3, :4] = value

            elif value.shape == (3, 3):
                self.rotation = value

            elif value.shape == (3,):
                self.translation = value

    @staticmethod
    def from_Rt(R, t):
        if isinstance(R, Quaternion):
            R = R.rotation_matrix
        return Transform(np.block([[R, np.mat(t).T], [np.zeros(3), 1]]))

    @staticmethod
    def from_euler(angles, axes='sxyz', degrees=False):
        if degrees:
            angles = np.deg2rad(angles)
        return Transform(t3d.euler.euler2mat(*angles, axes=axes))

    @staticmethod
    def from_transformed_points(A, B):
        assert A.shape == B.shape

        mean_A = np.mean(A, axis=0)
        mean_B = np.mean(B, axis=0)
        centroid_A = A - mean_A
        centroid_B = B - mean_B

        C = centroid_A.T @ centroid_B
        V, S, W = np.linalg.svd(C)

        if (np.linalg.det(V) * np.linalg.det(W)) < 0.0:
            S[-1] = -S[-1]
            V[:, -1] = -V[:, -1]

        R = V @ W
        t = mean_B - mean_A @ R

        return Transform.from_Rt(R.T, t)

    @staticmethod
    def random():
        return Transform.from_Rt(Quaternion.random(), np.random.rand(3))

    @property
    def rotation(self):
        return self.matrix[:3, :3]

    @property
    def quaternion(self):
        return Quaternion(t3d.quaternions.mat2quat(self.matrix[:3, :3]))

    @rotation.setter
    def rotation(self, rotation):
        if isinstance(rotation, Quaternion):
            rotation = rotation.rotation_matrix
        self.matrix[:3, :3] = rotation

    @property
    def position(self):
        return self.matrix[:3, 3].flatten()

    @position.setter
    def position(self, position):
        self.matrix[:3, 3] = np.array(position).reshape(3)

    @property
    def translation(self):
        return self.matrix[:3, 3].flatten()

    @translation.setter
    def translation(self, translation):
        self.matrix[:3, 3] = np.array(translation).reshape(3)

    @property
    def euler_angles(self, axes='sxyz'):
        return t3d.euler.mat2euler(self.matrix, axes=axes)

    @property
    def euler_degrees(self, axes='sxyz'):
        return np.rad2deg(t3d.euler.mat2euler(self.matrix, axes=axes))

    @property
    def T(self):
        return Transform(self.matrix.T)

    @property
    def inverse(self):
        return Transform.from_Rt(self.rotation.T, np.dot(-self.rotation.T, self.translation))

    def apply(self, points):
        points_4d = np.hstack([points[:, :3], np.ones((points.shape[0], 1))])
        transformed_4d = points_4d.dot(self.matrix.T)
        return np.hstack([transformed_4d[:, :3], points[:, 3:]])

    def interpolate(self, other, factor):
        assert (0 <= factor <= 1.0)
        other = Transform(other)
        return self.from_Rt(
            Quaternion.slerp(self.quaternion, other.quaternion, factor),
            self.position + factor * (other.position - self.position)
        )

    def __array__(self):
        return self.matrix

    def __getitem__(self, values):
        return self.matrix.__getitem__(values)

    def __add__(self, other):
        return Transform(other) @ self

    def __matmul__(self, other):
        if isinstance(other, np.ndarray):
            return self.apply(other)
        return Transform(self.matrix @ Transform(other).matrix)

    def __eq__(self, other):
        return np.allclose(self.matrix, other.matrix)

    def __repr__(self):
        return 'R=%s t=%s' % (
            np.array_str(self.euler_degrees, precision=3, suppress_small=True),
            np.array_str(self.position, precision=3, suppress_small=True)
        )



import cv2
import numpy as np
from PIL import Image, ImageDraw
import random

from .color_utils import map_colors
from .transform import Transform


class LidarCamera:
    world2cam = Transform.from_euler([-90, 0, -90], degrees=True)

    def __init__(self, camera_id):
        self.id = camera_id
        self.pose = Transform(self.world2cam)
        self.K = np.eye(3, dtype=np.float32)
        self.D = np.zeros(6, dtype=np.float32)
        self.model = 'brown_conrady'
        self.model = None
        self.scale_factor = 1
        self.skew = 0

    @property
    def position(self) -> np.ndarray:
        return self.pose.position

    @property
    def rotation(self) -> np.ndarray:
        return self.pose.rotation

    @property
    def world_transform(self) -> Transform:
        return self.pose @ self.world2cam.T

    @property
    def fx(self):
        return self.K[0, 0]

    @property
    def fy(self):
        return self.K[1, 1]

    @property
    def cx(self):
        return self.K[0, 2]

    @property
    def cy(self):
        return self.K[1, 2]

    @property
    def intrinsic_matrix(self):
        return self.K

    @property
    def extrinsic_matrix(self):
        return self.pose.inverse[:3, :4]

    @property
    def projection_matrix(self):
        return self.K @ self.extrinsic_matrix

    @position.setter
    def position(self, position: np.ndarray):
        self.pose.position = position

    @rotation.setter
    def rotation(self, rotation):
        self.pose.rotation = Transform(rotation).rotation

    @world_transform.setter
    def world_transform(self, transform: Transform):
        self.pose = transform @ self.world2cam

    @extrinsic_matrix.setter
    def extrinsic_matrix(self, matrix):
        self.pose = Transform(matrix).inverse

    @projection_matrix.setter
    def projection_matrix(self, P):
        assert P.shape == (3, 4), 'Projection matrix should be 3x4'
        K, R, t, _, _, _, _ = cv2.decomposeProjectionMatrix(P)
        self.pose = Transform.from_Rt(R.T, t[:3, 0] / t[3, 0])
        self.K = K

    def calibrate(self, position=None, rotation=None, pose=None, extrinsic_matrix=None, projection_matrix=None, K=None, D=None, model=None, scale_factor=None, skew=None, **kwargs):
        """
        Helper for camera calibration
        :param position: Camera position [x, y, z]
        :param rotation: Camera rotation/heading
        :param pose: Camera pose (position + rotation)
        :param extrinsic_matrix: extrinsic 4x4 matrix (world to camera transform)
        :param projection_matrix: 3x4 projection matrix
        :param K: intrinsic 3x3 matrix
        :param D: distortion values
        :param model: camera model
        :param scale_factor: image scale_factor
        :param skew: camera skew coefficient
        """
        if position is not None:
            self.position = position
        if rotation is not None:
            self.rotation = rotation
        if pose is not None:
            self.pose = Transform(pose)
        if extrinsic_matrix is not None:
            self.extrinsic_matrix = extrinsic_matrix
        if projection_matrix is not None:
            self.projection_matrix = projection_matrix
        if K is not None:
            self.K = np.array(K[:3, :3])
        if D is not None:
            self.D = np.pad(D, (0, 6 - len(D)), 'constant')
        if model is not None:
            self.model = model
        if scale_factor is not None:
            self.scale_factor = scale_factor
        if skew is not None:
            self.skew = skew
        if 'world_transform' in kwargs:
            self.world_transform = kwargs['world_transform']
        if 'fx' in kwargs:
            self.K[0, 0] = kwargs['fx']
        if 'fy' in kwargs:
            self.K[1, 1] = kwargs['fy']
        if 'cx' in kwargs:
            self.K[0, 2] = kwargs['cx']
        if 'cy' in kwargs:
            self.K[1, 2] = kwargs['cy']

    def apply_transform(self, transform: Transform):
        self.pose = transform @ self.pose

    def rotate(self, angles, degrees=True):
        self.apply_transform(Transform.from_euler(angles, degrees=degrees))

    def translate(self, vector):
        self.apply_transform(Transform(vector))

    def project_points(self, points: np.ndarray, use_distortion=False):
        projected = Transform(self.projection_matrix) @ points[:, :3]

        # projected = ((points[:, :3] - self.position) @ self.rotation) @ self.intrinsic_matrix[:3, :3].T
        projected[:, 0] /= np.where(projected[:, 2] == 0, np.inf, projected[:, 2])
        projected[:, 1] /= np.where(projected[:, 2] == 0, np.inf, projected[:, 2])

        if use_distortion:
            projected[:, :2] = cv2.fisheye.projectPoints(
                objectPoints=np.array([points[:, :3]], dtype=np.float32),
                rvec=cv2.Rodrigues(self.extrinsic_matrix[:3, :3])[0],
                tvec=self.extrinsic_matrix[:3, 3],
                D=np.array([self.D[0], self.D[1], self.D[4], self.D[5]], dtype=np.float32),
                K=np.array(self.K, dtype=np.float32),
                alpha= self.skew
            )[0].reshape((-1, 2))

        return np.hstack([
            projected[:, :3],
            points[:, 3:]
        ])

    def get_projected_image(self, image, points, frame_transform, color_mode='default', oversample=3):
        assert image, 'No image loaded.'
        def crop_points(points, bounding_box):
            conditions = np.logical_and(points[:, :3] >= bounding_box[0], points[:, :3] < bounding_box[1])
            mask = np.all(conditions, axis=1)
            return points[mask]
        im = image.get_image().convert('RGBA')
        radius = 3
        points = np.array(random.sample(points.tolist(), int(len(points) / 2))) #reduce the number of points projected, no need to project everything
        # Project points image
        points_im = Image.new('RGBA', (im.size[0] * oversample, im.size[1] * oversample))

        draw = ImageDraw.Draw(points_im)

        if self.model == 'cylindrical':
            # handle cylindrical cameras
            epsilon = 0.0000001
            fisheye = frame_transform.inverse @ np.array([points[:,0],points[:,1], points[:,2] ], dtype=np.float)  # 3D point in camera coordinates
            fisheye = fisheye.T  # 3D point in camera coordinates
            fisheye = self.pose.inverse @ fisheye
            fisheye[:, 1] *= -1  # invert y because cylindrical y is up and cartesian y is down
            fisheye = Transform(self.extrinsic_matrix[:3, :3]) @ fisheye.T  # lift cylinder to stand up straight
            cylindrical = np.array([np.arctan2(fisheye[0, :], fisheye[2, :]),
                                    fisheye[1, :] / np.sqrt(fisheye[0, :] ** 2 + fisheye[2, :] ** 2),
                                    np.ones(fisheye.shape[1])])
            cylindrical[1, :] *= -1   # invert y because cylindrical y is up and cartesian y is down
            q = self.K @ cylindrical
            img_coords = q[[0, 1], :]  # pixels on image
            img_coords = img_coords.T

            for point in img_coords[:, :2]:
                draw.ellipse([tuple(point - radius), tuple(point + radius)], fill=tuple([255, 10, 10]))
            points_im = points_im.resize(im.size, Image.CUBIC)

            # Merge images
            projected_im = Image.composite(points_im, im, points_im)
            return projected_im
        if self.model == 'fisheye':
            wct = self.pose @ frame_transform.T
            projected = self.project_points(points, use_distortion=True)
            #  projected = crop_points(projected,
                                    #  np.array([[0, 0, 0.1], [im.size[0], im.size[1], np.inf]]))
            fisheye = self.world_transform @ np.array([projected[:,0],projected[:,1], projected[:,2] ], dtype=np.float)  # 3D point in camera coordinates
            fisheye = fisheye.T  # 3D point in camera coordinates
            fisheye = np.concatenate((np.array(fisheye),np.array(projected[:,3])[:,None]),axis=1)

            if not len(fisheye):
                return im

            colors = map_colors(fisheye, color_mode)

            for point, color in zip(fisheye[:, :2] * oversample, colors):
                draw.ellipse([tuple(point - radius), tuple(point + radius)], fill=tuple(color))
            points_im = points_im.resize(im.size, Image.CUBIC)

            # Merge images
            projected_im = Image.composite(points_im, im, points_im)
            return projected_im
        else:
            projected = self.project_points(points, use_distortion=False)
            projected = crop_points(projected,
                                    np.array([[0, 0, 0.1], [im.size[0], im.size[1], np.inf]]))

            # Returns original image if not projected points on image
            if not len(projected):
                return im

            colors = map_colors(projected, color_mode)

            for point, color in zip(projected[:, :2] * oversample, colors):
                draw.ellipse([tuple(point - radius), tuple(point + radius)], fill=tuple(color))
            points_im = points_im.resize(im.size, Image.CUBIC)

            # Merge images
            projected_im = Image.composite(points_im, im, points_im)
            return projected_im

    def __repr__(self):
        return 'LidarCamera({0}) {1}'.format(self.id, self.pose)



import cv2
import numpy as np
from matplotlib.colors import hsv_to_rgb


def normalize(values):
    return np.interp(values, [np.percentile(values, 5), np.percentile(values, 95)], [0, 1])


def map_colors(points, mode='default', color_map=cv2.COLORMAP_JET):

    if mode == 'depth':
        values = np.uint8(normalize(points[:, 2]) * 255)
        rgb = cv2.applyColorMap(values, color_map)
        return rgb.reshape((-1, 3))

    if mode == 'intensity':
        values = 255 - np.uint8(normalize(points[:, 3]) * 255)
        rgb = cv2.applyColorMap(values, color_map)
        return rgb.reshape((-1, 3))

    if mode == 'default':
        x, y, z, i = points.T
        i_range = [np.percentile(i, 10), np.percentile(i, 90)]

        value = np.sum([
            np.interp(z, [0, 50], [100, -10]),
            np.interp(i, i_range, [-10, 10])
        ], axis=0)

        rgb = hsv_to_rgb(np.array([
            np.interp(value, [0, 100], [220 / 360, 1 / 360]),
            np.interp(i, i_range, [0.5, 1]),
            np.interp(i, i_range, [0.4, 1]),
        ]).T)

        alpha = np.interp(i, [np.percentile(i, 10), np.percentile(i, 90)], [0.75, 1])

        return np.uint8(np.hstack([
            rgb.reshape((-1, 3)),
            alpha.reshape((-1, 1))
        ]) * 255)

    # other mode
    values = np.uint8(normalize(points[:, 2]) * 255)
    rgb = cv2.applyColorMap(values, color_map)
    alpha = np.uint8(normalize(points[:, 3]) * 255)
    return np.hstack([rgb.reshape((-1, 3)), alpha.reshape((-1, 1))])


import hashlib
import os
import shutil
import tempfile
import zipfile
from concurrent import futures
from io import BytesIO
from typing import Any, MutableMapping, List, Dict

import boto3
import numpy as np
import pandas as pd
import ujson
from pyquaternion import Quaternion
from PIL import Image, ImageEnhance
from scaleapi import Task
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, Button
import matplotlib.gridspec as gridspec

from .camera import LidarCamera
from .transform import Transform
from .connectors import Importer

UPLOAD_POOL_SIZE = 8
s3 = boto3.client('s3')


def format_lidar_point(p):
    return dict(zip(('x', 'y', 'z', 'i', 'd'), p))


def format_point(p):
    return dict(zip(('x', 'y', 'z'), p))


def format_quaternion(q):
    return dict(zip(['w', 'x', 'y', 'z'], q))


def fp_md5(fp: BytesIO):
    fp.seek(0)
    md5 = hashlib.md5()
    buf = fp.read(2 ** 20)
    while buf:
        md5.update(buf)
        buf = fp.read(2 ** 20)
    fp.seek(0)
    return md5.hexdigest()


def get_s3_etag(bucket, key):
    try:
        return s3.head_object(Bucket=bucket, Key=key)['ETag'].strip('"')
    except Exception:
        pass


def s3_smart_upload(bucket, key, fileobj, content_type):
    s3_hash = get_s3_etag(bucket, key)
    local_hash = fp_md5(fileobj)

    if s3_hash == local_hash:
        print(f'File exists: {bucket}/{key}')
        return

    print(f'Uploading {bucket}/{key}...')
    s3.upload_fileobj(
        Fileobj=fileobj,
        Bucket=bucket,
        Key=key,
        ExtraArgs={'ContentType': content_type}
    )


class LidarImage:
    """LidarImage objects are represent an image and LidarCamera reference."""

    def __init__(self, camera):
        self.camera = camera
        self.file = None
        self.transform = None
        self.metadata = None

    def load_file(self, file):
        if not isinstance(file, str):
            print('WARNING: No file!')
        self.file = file

    def load_pil_image(self, pil_image: Image.Image):
        self.file = tempfile.mktemp(suffix='jpg')
        pil_image.save(self.file, format='JPEG', quality=70, optimize=True)
        print(f'Temp file created: {self.file}')

    # Legacy method
    def load(self, image):
        print('DEPRECATED METHOD LidarImage.load(), please use .load_file() or .load_pil_image()')
        if isinstance(image, Image.Image):
            self.load_pil_image(image)
        else:
            self.load_file(image)

    def get_file(self):
        return self.file

    def get_image(self):
        return Image.open(self.file)

    def scale(self, scale_factor):
        im = self.get_image()
        size = (int(im.width * scale_factor), int(im.height * scale_factor))
        self.load_pil_image(im.resize(size, Image.LANCZOS))

    def as_array(self):
        return np.asarray(self.get_image())

    def set_brightness(self, factor):
        im = ImageEnhance.Brightness(self.get_image()).enhance(factor)
        self.load_pil_image(im)

    def save(self, target_file):
        shutil.copyfile(self.file, target_file)

    def s3_upload(self, bucket, key):
        with open(self.get_file(), 'rb') as fp:
            s3_smart_upload(
                bucket=bucket,
                key=key,
                fileobj=fp,
                content_type='image/jpeg'
            )


class LidarFrame:
    """Frame objects represent all the point cloud, image, and other data that is sent to the annotator."""

    def __init__(self, frame_id, cameras):
        self.id = frame_id
        self.cameras: pd.Series[Any, LidarCamera] = cameras
        self.images: pd.Series[Any, LidarImage] = pd.Series(dtype=object)
        self.points: np.ndarray = np.zeros((0, 5), dtype=float)
        self.transform = Transform()

    def get_image(self, camera_id) -> LidarImage:
        """
        Get a image by camera_id and creates one if not exists
        :param camera_id: The camera id
        :return: LidarImage
        """
        assert camera_id in self.cameras, 'Camera not found'
        if camera_id not in self.images:
            if isinstance(camera_id, int):
                self.images.index = self.images.index.astype(int)
            self.images[camera_id] = LidarImage(camera=self.cameras[camera_id])
        return self.images[camera_id]

    def add_points_from_connector(self, connector: Importer, transform: Transform = None, intensity=1, sensor_id=0):
        self.add_points(connector.data, transform, intensity, sensor_id)

    def add_points(self, points: np.array, transform: Transform = None, intensity=1, sensor_id=0):
        if points.ndim == 1:
            points = np.array([points])
        if points.shape[1] == 3:
            points = np.hstack([points, np.ones((points.shape[0], 1)) * intensity])
        if transform is not None:
            points = transform.apply(points)

        points = np.hstack([points, np.ones((points.shape[0], 1)) * sensor_id])

        self.points = np.vstack([self.points, points])

    def add_debug_lines(self, intensity=1, length=5, device=0):
        x_line = np.array([np.array([length * k / 100, 0, 0]) for k in range(0, 100)])
        for camera in self.cameras:
            self.add_points(x_line, transform=camera.world_transform, intensity=intensity)

    def get_world_points(self):
        return np.hstack([self.transform @ self.points[:, :3], self.points[:, 3:4], self.points[:, 4:5]])

    # leeaving this method as legacy / old code dependency
    def get_projected_image(self, camera_id, color_mode='default', **kwargs):
        return self.cameras[camera_id].get_projected_image(self.get_image(camera_id), self.points, self.transform, color_mode, **kwargs)

    def manual_calibration(self, camera_id, intrinsics_ratio=1000, extrinsics_ratio=10):
        fig = plt.figure(constrained_layout=True)
        intrinsics = ['fx','fy','cx', 'cy']
        extrinsics_position = ['x','y','z']
        extrinsics_heading = ['qw','qx','qy','qz']
        heights = np.concatenate((np.array([7]), np.array([0.2 for x in range(0, 7)])), axis=0)
        gs = fig.add_gridspec(ncols= 3, nrows=1 + len(extrinsics_heading) + len(extrinsics_position), height_ratios=heights)
        imgObj = fig.add_subplot(gs[0, :])
        imgObj.imshow(self.cameras[camera_id].get_projected_image(self.get_image(camera_id), self.points, self.transform, 'depth', 1))

        for index, key in enumerate(intrinsics):
            globals()[f"ax{key}"] = fig.add_subplot(gs[index + 1, 0])
            value = getattr(self.cameras[camera_id], key)
            globals()[key] = Slider(globals()[f"ax{key}"], f"{key}", value-intrinsics_ratio, value+intrinsics_ratio, valinit=value)

        for index, key in enumerate(extrinsics_position):
            globals()[f"ax{key}"] = fig.add_subplot(gs[index + 1, 1])
            value = getattr(self.cameras[camera_id], 'position')
            globals()[key] = Slider(globals()[f"ax{key}"], f"{key}", value[index]-extrinsics_ratio, value[index]+extrinsics_ratio, valinit=value[index])

        for index, key in enumerate(extrinsics_heading):
            globals()[f"ax{key}"] = fig.add_subplot(gs[index + len(extrinsics_position) + 1, 1])
            value = getattr(self.cameras[camera_id], 'rotation')
            value = Quaternion(matrix=value)
            globals()[key] = Slider(globals()[f"ax{key}"], f"{key}", value[index]-extrinsics_ratio, value[index]+extrinsics_ratio, valinit=value[index])

        def update(val):
            self.cameras[camera_id].calibrate(K=np.array([
                [fx.val, 0, cx.val],
                [0, fy.val, cy.val],
                [0,0,1]
                ]),
                pose= Transform.from_Rt(
                    Quaternion(qw.val,qx.val,qy.val,qz.val).rotation_matrix,
                    [x.val,y.val,z.val]
                    )
                )
            np.set_printoptions(suppress=True)
            print(f"New intrinsics for Camera {camera_id}")
            print(self.cameras[camera_id].K)
            print(f"New position for Camera {camera_id}")
            print(self.cameras[camera_id].position)
            imgObj.imshow(self.cameras[camera_id].get_projected_image(self.get_image(camera_id), self.points, self.transform, 'depth', 1))
            fig.canvas.draw_idle()

        axApply = fig.add_subplot(gs[1, 2])
        bApply = Button(axApply, 'Apply changes')
        bApply.on_clicked(update)
        plt.show()
    def get_filename(self):
        return "frame-%s.json" % self.id

    def apply_transform(self, T):
        self.transform = Transform(T) @ self.transform

    def filter_points(self, min_intensity=None, min_intensity_percentile=None):
        if min_intensity is not None:
            self.points = self.points[self.points[:, 3] >= min_intensity]
        if min_intensity_percentile is not None:
            self.points = self.points[self.points[:, 3] >= np.percentile(self.points[:, 3], min_intensity_percentile)]

    def to_json(self, public_url=''):

        def format_image(camera):
            image = self.images[camera.id]
            wct = (image.transform or self.transform) @ camera.pose
            result = dict(
                position=format_point(wct.position),
                heading=format_quaternion(wct.quaternion),
                image_url='%s/image-%s-%s.jpg' % (public_url, camera.id, self.id),
                camera_model=camera.model,
                fx=camera.fx,
                fy=camera.fy,
                cx=camera.cx,
                cy=camera.cy,
                skew=camera.skew,
                k1=float(camera.D[0]),
                k2=float(camera.D[1]),
                p1=float(camera.D[2]),
                p2=float(camera.D[3]),
                k3=float(camera.D[4]),
                k4=float(camera.D[5]),
                scale_factor=camera.scale_factor
            )
            if image.metadata:
                result['metadata'] = image.metadata
            return result

        images_json = self.cameras[self.images.index].apply(format_image).to_json(orient='records')
        points_json = pd.DataFrame(self.get_world_points(), columns=['x', 'y', 'z', 'i', 'd']) \
            .to_json(double_precision=4, orient='records', date_format=None)

        out = ujson.dumps({
            'images': "__IMAGES__",
            'points': "__POINTS__",
            'device_position': format_point(self.transform.position),
            'device_heading': format_quaternion(self.transform.quaternion)
        })

        out = out.replace('"__IMAGES__"', images_json)
        out = out.replace('"__POINTS__"', points_json)

        return out

    def save(self, path, public_url=''):
        # Save frame
        with open(os.path.join(path, 'frame-%s.json' % self.id), 'w') as file:
            file.write(self.to_json(public_url))

        # Save images
        for camera_id, image in self.images.items():
            image.save(os.path.join(path, 'image-%s-%s.jpg' % (camera_id, self.id)))

    def s3_upload(self, bucket, path):
        # print(f'Uploading frame {self.id}...')
        public_url = f"s3://{bucket}/{path}"

        # Upload frame json file
        s3_smart_upload(
            fileobj=BytesIO(bytes(self.to_json(public_url), encoding='utf-8')),
            bucket=bucket,
            key=f'{path}/frame-{self.id}.json',
            content_type='application/json'
        )

        # Upload images
        for camera_id, image in self.images.items():
            image.s3_upload(bucket, f'{path}/image-{camera_id}-{self.id}.jpg')

    def __repr__(self):
        return 'Frame({0}) {1}'.format(self.id, self.transform)


class LidarScene:
    """LidarScene object represent all frames on a scene."""

    def __init__(self):
        """

        :rtype: object
        """
        self.cameras: MutableMapping[LidarCamera] = pd.Series()
        self.frames: MutableMapping[LidarFrame] = pd.Series()
        self.public_url = None

    def get_camera(self, camera_id=None, index=None) -> LidarCamera:
        """
        Get a camera by id (or index) and creates one if not exists
        :param index: The camera index
        :param camera_id: The camera id
        :return: LidarCamera
        """
        assert camera_id is not None or index is not None, 'id or index must be specified'

        if camera_id is None:
            camera_id = self.cameras.index[index]

        if camera_id not in self.cameras:
            if isinstance(camera_id, int):
                self.cameras.index = self.cameras.index.astype(int)
            self.cameras[camera_id] = LidarCamera(camera_id)
        return self.cameras[camera_id]

    def get_frame(self, frame_id=None, index=None) -> LidarFrame:
        """
        Get a frame by id (or index) and creates one if not exists
        :param index: The frame index
        :param frame_id: The frame id
        :return: LidarFrame
        """
        assert frame_id is not None or index is not None, 'id or index must be specified'

        if frame_id is None:
            frame_id = self.frames.index[index]

        if frame_id not in self.frames:
            if isinstance(frame_id, int):
                self.frames.index = self.frames.index.astype(int)
            self.frames[frame_id] = LidarFrame(frame_id, cameras=self.cameras)
        return self.frames[frame_id]

    def apply_transforms(self, world_transforms: List[Transform]):
        for idx in range(len(self.frames)):
            self.get_frame(index=idx).apply_transform(world_transforms[idx])

    def filter_points(self, min_intensity=None, min_intensity_percentile=None):
        for frame in self.frames:
            frame.filter_points(min_intensity, min_intensity_percentile)

    def get_projected_image(self, camera_id, color_mode='intensity', frames_index=range(0,1), **kwargs):
        all_points = np.array([]).reshape(0,4)
        for idx in frames_index:
            points = np.array(self.frames[idx].points)[:,:4]
            print(points.shape)
            points[:, 3] = idx
            all_points = np.concatenate((all_points, points), axis=0)
        return self.cameras[camera_id].get_projected_image(self.frames[frames_index[0]].get_image(camera_id), all_points, self.frames[frames_index[0]].transform, color_mode)

    def apply_transform(self, world_transform: Transform):
        for idx in range(len(self.frames)):
            self.get_frame(index=idx).apply_transform(world_transform)

    def make_transforms_relative(self):
        offset = self.get_frame(index=0).transform.inverse
        for frame in self.frames:
            frame.transform = offset @ frame.transform

    def to_dict(self, public_url=None):
        if public_url is None:
            public_url = self.public_url
        return dict(frames=['%s/frame-%s.json' % (public_url, frame.id) for frame in self.frames])

    def s3_upload(self, bucket, path=None, mock_upload=False, use_threads=True):
        self.public_url = f"s3://{bucket}/{path}"

        print('Uploading scene to S3: %s' % self.public_url)
        scene_dict = self.to_dict(self.public_url)

        poses_csv = pd.DataFrame(self.frames.map(lambda f: list(f.transform.matrix.reshape(-1))).to_dict()).T.to_csv(
            header=False)

        if not mock_upload:
            # Upload scene json file
            s3_smart_upload(
                bucket=bucket,
                key=f'{path}/scene.json',
                fileobj=BytesIO(bytes(ujson.dumps(scene_dict), encoding='utf-8')),
                content_type='application/json'
            )

            # Upload ego2world csv file
            s3_smart_upload(
                bucket=bucket,
                key=f'{path}/ego2world.csv',
                fileobj=BytesIO(bytes(poses_csv, encoding='utf-8')),
                content_type='text/plain'
            )

            if use_threads:
                with futures.ThreadPoolExecutor(max_workers=UPLOAD_POOL_SIZE) as executor:
                    future_list = [
                        executor.submit(LidarFrame.s3_upload, frame, bucket, path)
                        for frame in self.frames
                    ]

                    for future in future_list:
                        future.result()

            else:
                for frame in self.frames:
                    frame.s3_upload(bucket, path)

        signed_url = s3.generate_presigned_url('get_object', Params={
            'Bucket': bucket,
            'Key': f'{path}/scene.json'
        })

        print(f'Scene uploaded: {signed_url}')
        return self.public_url

    def save_task(self, filepath, template=None):
        print('Saving scene:', filepath)
        with zipfile.ZipFile(filepath, mode='w') as out:
            # Save task
            scene_dict = self.to_dict()
            task_dict = dict(
                template or {},
                attachment_type='json',
                attachments=scene_dict['frames']
            )
            out.writestr('task.json', ujson.dumps(task_dict))

            # Save frames
            for frame in self.frames:
                # Save points
                data = frame.to_json(self.public_url)
                out.writestr('frame-%s.json' % frame.id, data, compress_type=zipfile.ZIP_DEFLATED)

                # Save frame images
                for camera_id, image in frame.images.items():
                    if not image.file:
                        assert NotImplemented('Only file-imported Images supported')
                    out.write(image.file, 'image-%s-%s.jpg' % (camera_id, frame.id))
        print('Scene saved.')

    def create_task(self, template: Dict = None, task_type: str = 'lidarannotation') -> Task:
        """ Creates a Scale platform task from the configured scene

        Args:
            template: Dictionary of payload for task creation
            task_type: Select a Scale API endpoint top upload data to, currently supports 'lidarannotation' and 'lidartopdown'. Defaults to 'lidarannotation'.

        Returns:
            Task object with related information. Inherited from `scaleapi.Task` object.
        """
        if task_type == 'lidarannotation':
            from .task import LidarAnnotationTask
            return LidarAnnotationTask.from_scene(self, template)
        elif task_type == 'lidartopdown':
            from .task import LidarTopDownTask
            return LidarTopDownTask.from_scene(self, template)
        else:
            assert NotImplemented(f'Specified task_type {task_type} is not supported')



import os

import numpy as np
import requests
import ujson
from scaleapi import ScaleClient, Task

from .scene import LidarScene


def parse_xyz(obj):
    return np.array([obj['x'], obj['y'], obj['z']])


def parse_cuboid_frame(frame):
    return np.array([parse_xyz(p) for p in [c['position'] for c in frame['cuboids']]])


def read_cuboid_points(filename):
    # Read from result file
    with open(filename, 'r') as json_file:
        return np.array([parse_cuboid_frame(frame) for frame in ujson.load(json_file)])


def get_api_client() -> ScaleClient:
    assert os.getenv('SCALE_API_KEY'), 'No API key, please set SCALE_API_KEY environment variable'
    return ScaleClient(os.getenv('SCALE_API_KEY'))


def get_default_template():
    return {
        'task_id': None,
        'callback_url': 'http://example.com',
        'instruction': 'Please label all objects',
        'labels': [
            'Object'
        ],
        'meters_per_unit': 1,
        'max_distance_meters': 20
    }


class LidarAnnotationTask(Task):
    scene: LidarScene = None

    def __init__(self, param_dict, client):
        super(LidarAnnotationTask, self).__init__(param_dict, client)

    @staticmethod
    def from_scene(scene: LidarScene, template=None, client=None):
        assert scene.public_url, 'No public URL on scene, please upload or save the scene first'

        # Get client using environ api key
        if client is None:
            client = get_api_client()

        # Starting with default template as params
        param_dict = get_default_template()

        # Load scene params
        scene_dict = scene.to_dict(scene.public_url)
        param_dict['attachments'] = scene_dict['frames']
        param_dict['attachment_type'] = 'json'

        if isinstance(template, dict):
            param_dict.update(template)

        elif isinstance(template, str):
            param_dict.update(ujson.load(open(template)))

        elif template is not None:
            raise AttributeError('Template error')

        return LidarAnnotationTask(param_dict, client)

    @staticmethod
    def from_id(task_id):
        task = get_api_client().fetch_task(task_id)
        return LidarAnnotationTask(task.param_dict, task.client)

    def get_annotations(self):
        assert 'response' in self.param_dict, 'Task without response'
        url = self.param_dict['response']['annotations']['url']
        response = requests.get(url)
        return ujson.loads(response.text)

    def get_cuboid_positions_by_frame(self):
        annotations = self.get_annotations()
        return np.array([
            np.array([
                parse_xyz(p)
                for p in [c['position'] for c in frame['cuboids']]
            ])
            for frame in annotations
        ])

    def publish(self, task_type='lidarannotation'):
        task = self.client.create_task(task_type, **self.param_dict)
        print('Task created: %s' % task)
        return task


class LidarTopDownTask(Task):
    scene: LidarScene = None

    def __init__(self, param_dict, client):
        super(LidarTopDownTask, self).__init__(param_dict, client)

    @staticmethod
    def from_scene(scene: LidarScene, template=None, client=None):
        assert scene.public_url, 'No public URL on scene, please upload or save the scene first'

        # Get client using environ api key
        if client is None:
            client = get_api_client()

        # Starting with default template as params
        param_dict = get_default_template()

        # Load scene params
        scene_dict = scene.to_dict(scene.public_url)
        param_dict['attachments'] = scene_dict['frames']
        param_dict['attachment_type'] = 'json'

        if isinstance(template, dict):
            param_dict.update(template)

        elif isinstance(template, str):
            param_dict.update(ujson.load(open(template)))

        elif template is not None:
            raise AttributeError('Template error')

        return LidarTopDownTask(param_dict, client)

    @staticmethod
    def from_id(task_id):
        task = get_api_client().fetch_task(task_id)
        return LidarTopDownTask(task.param_dict, task.client)

    def publish(self, task_type='lidartopdown'):
        task = self.client.create_task(task_type, **self.param_dict)
        print('Task created: %s' % task)
        return task



from abc import ABCMeta, abstractmethod

import numpy as np
import open3d as o3d
import pandas as pd
from laspy.file import File

_FIELDS = ['x', 'y', 'z', 'i', 'd']


class Importer:
    """Namespace for all import connectors

    """
    class Base(metaclass=ABCMeta):
        """Abstract importer class to be inherited for data type specific implementation

        """

        def __init__(self, fp: str) -> None:
            """ Constructor to be called for preparing filepaths

            Args:
                fp: Relative or absolute path to file to be loaded in explicit `load` method
            """
            self._filepath: str = fp
            self._data: pd.DataFrame = pd.DataFrame(
                columns=_FIELDS
            )

        def load(self, **kwargs) -> None:
            self._load(**kwargs)

        @property
        def data(self):
            return self._data.to_numpy()

        @abstractmethod
        def _load(self, **kwargs) -> None:
            ...

    class CSV(Base, metaclass=ABCMeta):
        """Metaclass for specific CSV Importer implementations.
        Due to non-standardized csv format, it is recommended to write more specific implementations on a case-by-case basis.
        """

        @abstractmethod
        def _load(self, **kwargs):
            ...

    class OrderedCSV(CSV):
        """Expects a csv file with or without header, but assumes correct order of columns / values as follows: `[x, y, z, i, d]`
        Cuts off any exceeding column count.
        """

        def _load(self, **kwargs):
            self._data = pd.read_csv(self._filepath, **kwargs)
            c_count = len(self._data.columns)
            if c_count > 5:  # Remove columns exceeding count of FIELDS
                self._data = self._data.drop(
                    self._data.columns[5:],
                    axis=1
                )
                c_count = 5

            self._data.columns = _FIELDS[:c_count]

    class NamedCSV(CSV):
        """Expects a csv file with header row and column names matching `x`, `y`, `z`, `i`, `d`
        Dismisses any columns not matching any of the expected names. Case-sensitive.
        """

        def _load(self, **kwargs):
            self._data = pd.read_csv(self._filepath, **kwargs)
            exists = []
            for c in _FIELDS:
                if c in self._data.columns:
                    exists.append(c)
            self._data = self._data[exists]

    class PCD(Base):
        """Uses open3d library to read pcd files, expects `[x, y, z]`, dismisses other columns.

        """

        def _load(self, **kwargs):
            self._data = pd.DataFrame(
                np.asarray(o3d.io.read_point_cloud(self._filepath).points)[:, :3],
                columns=_FIELDS[:3]
            )

    class LAS(Base):
        """Uses Laspy library to read .las file and expects properties `x`, `y`, `z`, `intensity` to be present.

        """

        def _load(self, **kwargs):
            las_file = File(self._filepath, mode='r')
            self._data = pd.DataFrame(
                np.column_stack([
                    las_file.x,
                    las_file.y,
                    las_file.z,
                    las_file.intensity
                ]),
                columns=['x', 'y', 'z', 'i']
            )
